import pymongo
from datetime import datetime, timedelta
import traceback
from pymongo import ReturnDocument
from datetime import datetime, timedelta
import time


class Job:

    def __init__(self, queue, data):
        """
        """
        self._queue = queue
        self._data = data
    
    @property
    def payload(self):
        return self._data['payload']

    @property
    def job_id(self):
        return self._data["_id"]

    @property
    def priority(self):
        return self._data["priority"]

    @property
    def attempts(self):
        return self._data["attempts"]

    @property
    def locked_by(self):
        return self._data["locked_by"]

    @property
    def locked_at(self):
        return self._data["locked_at"]

    @property
    def queued_at(self):
        return self._data["queued_at"]

    @property
    def depends_on(self):
        return self._data["depends_on"]

    @property
    def last_error(self):
        return self._data["last_error"]
    
    @property
    def state(self):
        return self._data.get("state")

    @property
    def run_after(self):
        return self._data.get("run_after")

    @property
    def progress_count(self):
        return self._data.get("progress", 0)

    ## Job Controller

    def complete(self):
        """Job has been completed.
        """
        # remove the dependency from other jobs.
        deleted_job = self._queue.collection.delete_one(
            filter={"_id": self.job_id, "locked_by": self._queue.consumer_id})
        
        def remove_from_dependencies(tries=0):
            if tries >= 3:
                return
            try:
                self._queue.collection.update_many({"depends_on": self.job_id},
                                 {"$pull": {"depends_on": self.job_id}})
            except pymongo.errors.WriteError as e:
                time.sleep(0.5*tries)
                remove_from_dependencies(tries=tries+1)
        remove_from_dependencies()
        return deleted_job

    def error(self, message=None):
        """Note an error processing a job, and return it to the queue.
        """
        self._data = self._queue.collection.find_one_and_update(
            filter={"_id": self.job_id, "locked_by": self._queue.consumer_id},
            update={"$set": {
                "locked_by": None, "locked_at": None, "last_error": message},
                "$inc": {"attempts": 1}},
            return_document=ReturnDocument.AFTER)

    def progress(self, count=0):
        """Note progress on a long running task.
        """
        self._data = self._queue.collection.find_one_and_update(
            filter={"_id": self.job_id, "locked_by": self._queue.consumer_id},
            update={"$set": {"progress": count, "locked_at": datetime.now()}},
            return_document=ReturnDocument.AFTER)

    def release(self, sleep=0, state=None, inc_attempt=True):
        """Put the job back into_queue.
        """
        now = datetime.now()
        now_plus_seconds = now + timedelta(seconds = sleep)        
        attempt_increment_by = 1 if inc_attempt else 0
        self._data = self._queue.collection.find_one_and_update(
            filter={"_id": self.job_id, "locked_by": self._queue.consumer_id},
            update={"$set": {"locked_by": None, "locked_at": None, "run_after": now_plus_seconds, "state": state},
                    "$inc": {"attempts": attempt_increment_by}},
            return_document=ReturnDocument.AFTER)

    def __str__(self):
        return str(self._data)

    ## Context Manager support

    def __enter__(self):
        return self._data

    def __exit__(self, type, value, tb):
        if (type, value, tb) == (None, None, None):
            self.complete()
        else:
            error = traceback.format_exc()
            self.error(error)


import pymongo
from datetime import datetime, timedelta
from mongo_queue.job import Job
from uuid import uuid4
from pymongo import errors, ReturnDocument
from bson import ObjectId

DEFAULT_INSERT = {
    "attempts": 0,
    "locked_by": None,
    "locked_at": None,
    "last_error": None,
    "depends_on": []
}


class Queue:
    def __init__(self, collection, consumer_id, timeout=300, max_attempts=3):
        """
        """
        self.collection = collection
        self.consumer_id = consumer_id
        self.timeout = timeout
        self.max_attempts = max_attempts
        self.ensure_indexes()

    def __str__(self):
        return str({
            "consumer_id": self.consumer_id,
            "timeout": self.timeout,
            "max_attempts": self.max_attempts,
            "collection": self.collection
        })

    def ensure_indexes(self):
        """
        "locked_by": None,
                   "locked_at": None,
                   "channel": channel,
                   "attempts"
        """
        next_index = pymongo.IndexModel([("locked_by", pymongo.ASCENDING), ("locked_at", pymongo.ASCENDING),
                                         ("queued_at", pymongo.ASCENDING),
                                         ("channel", pymongo.ASCENDING),
                                         ("attempts", pymongo.ASCENDING),
                                         ("depends_on", pymongo.ASCENDING)], name="next_index")
        update_index = pymongo.IndexModel([("_id", pymongo.ASCENDING),
                                           ("locked_by", pymongo.ASCENDING)], name="update_index")
        depend_index = pymongo.IndexModel([("depends_on", pymongo.ASCENDING)], name="depend_index", sparse=True)
        unique_index = pymongo.IndexModel([("job_id", pymongo.ASCENDING),
                                           ("channel", pymongo.ASCENDING)], name="unique_index", unique=True)
        self.collection.create_indexes([next_index, update_index, unique_index, depend_index])

    def close(self):
        """Close the in memory queue connection.
        """
        self.collection.connection.close()

    def clear(self):
        """Clear the queue.
        """
        return self.collection.drop()

    def size(self):
        """Total size of the queue
        """
        return self.collection.count_documents(filter={})

    def repair(self):
        """Clear out stale locks.
        Increments per job attempt counter.
        """
        self.collection.find_one_and_update(
            filter={
                "locked_by": {"$ne": None},
                "locked_at": {
                    "$lt": datetime.now() - timedelta(seconds=self.timeout)}},
            update={
                "$set": {"locked_by": None, "locked_at": None},
                "$inc": {"attempts": 1}}
        )
        # TODO: Find the jobs with dependencies and the job it depends on does not exist.

    def drop_max_attempts(self):
        """
        """
        self.collection.update_many(filter={},
                                    update={"attempts": {"$gte": self.max_attempts}},
                                    remove=True)

    def put(self, payload, priority=0, channel="default", job_id=None, depends_on=[]):
        """Place a job into the queue
        """
        depends_on_bson = Queue._depends_on_bson(depends_on)
        job = dict(DEFAULT_INSERT)
        job['priority'] = priority
        job['payload'] = payload
        job['channel'] = channel
        job['job_id'] = job_id or str(uuid4())
        job['depends_on'] = depends_on_bson
        job['queued_at'] = datetime.now()
        try:
            return self.collection.insert_one(job).inserted_id
        except errors.DuplicateKeyError as e:
            return False

    def next(self, channel="default", recursion_limit: int = 10, current_level: int = 0):
        if current_level > recursion_limit:
            return None
        aggregate_result = list(self.collection.aggregate([
            {'$match': {'locked_by': None, 'locked_at': None,
                        "channel": channel,
                        "attempts": {"$lt": self.max_attempts},
                        "$or": [{"run_after": { "$exists": False }}, { "run_after" : {"$lt": datetime.now()}}]
            }},
            {"$lookup":
                {
                    "from": self.collection.name,
                    "localField": "depends_on",
                    "foreignField": "_id",
                    "as": "dependencies"
            }},
            {"$addFields": {
                "dependencies": {"$size": "$dependencies"}
            }},    
            {"$match": {"dependencies": {"$eq": 0}}},
            {"$sort": {'priority': pymongo.DESCENDING, "queued_at": pymongo.ASCENDING}},
            {"$limit": 1}
        ], allowDiskUse=True))
        if not aggregate_result:
            return None
        next_job = self.collection.find_one_and_update(
            filter={"_id": aggregate_result[0]["_id"], 'locked_by': None, 'locked_at': None},
            update={"$set": {"locked_by": self.consumer_id,
                             "locked_at": datetime.now()}},
            sort=[('priority', pymongo.DESCENDING)],
            return_document=ReturnDocument.AFTER
        )
        if not next_job:
            return self.next(channel=channel, recursion_limit=recursion_limit, current_level=current_level+1)
        else:
            next_job = self._wrap_one(next_job)

        return next_job

    def find_job_by_id(self, _id):
        if not _id:
            raise AttributeError("id required.")
        if not isinstance(_id, ObjectId):
            _id = ObjectId(_id)
        return self._wrap_one(self.collection.find_one(filter={"_id": _id}))

    def _jobs(self):
        return self.collection.find(
            query={"locked_by": None,
                   "locked_at": None,
                   "attempts": {"$lt": self.max_attempts}},
            sort=[('priority', pymongo.DESCENDING)],
        )

    def _wrap_one(self, data):
        return data and Job(self, data) or None

    @staticmethod
    def _depends_on_bson(depends_on):
        if not depends_on:
            return depends_on
        depends_on_bson = []
        for item in depends_on:
            if isinstance(item, ObjectId):
                depends_on_bson.append(item)
            elif isinstance(item, str):
                depends_on_bson.append(ObjectId(item))
            else:
                print("Unsupported dependency", item)
        return depends_on_bson




    def stats(self):
        """Get statistics on the queue.
        Use sparingly requires a collection lock.
        """

        js = """function queue_stat(){
        return db.eval(
        function(){
           var a = db.%(collection)s.count(
               {'locked_by': null,
                'attempts': {$lt: %(max_attempts)i}});
           var l = db.%(collection)s.count({'locked_by': /.*/});
           var e = db.%(collection)s.count(
               {'attempts': {$gte: %(max_attempts)i}});
           var t = db.%(collection)s.count();
           return [a, l, e, t];
           })}""" % {
            "collection": self.collection.name,
            "max_attempts": self.max_attempts}

        return dict(zip(
            ["available", "locked", "errors", "total"],
            self.collection.database.eval(js)))



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="mongo_queue_service",
    version="0.1.3",
    author="Amit Chotaliya",
    author_email="amit@shunyeka.com",
    description="Queue service built on top of mongo.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/shunyeka/mongo_queue/",
    download_url='https://github.com/shunyeka/mongo_queue/archive/v0.1.3.tar.gz',
    packages=setuptools.find_packages(),
    keywords=['mongo', 'queue', 'priority queue', 'task queue'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'pymongo'
    ],
    python_requires='>=3.6',
)


import os
import time

from datetime import datetime

import pymongo
from unittest import TestCase

from mongo_queue import Queue
from mongo_queue.lock import MongoLock, lock



class MongoLockTest(TestCase):

    def setUp(self):

        self.client = pymongo.MongoClient(os.environ.get("MONGO_URI"), retryWrites=False, maxPoolSize=1, directConnection=True)
        self.db = self.client.test_queue
        self.collection = self.db.locks

    def tearDown(self):
        self.client.drop_database("test_queue")

    def test_lock_acquire_release_context_manager(self):
        with lock(self.collection, 'test1') as l:
            self.assertTrue(l.locked)
        self.assertEqual(self.collection.count_documents(filter={}), 0)

    def test_auto_expires_old(self):
        lock = MongoLock(self.collection, 'test2', lease=2)
        self.assertTrue(lock.acquire())

        time.sleep(2.2)
        # Reports truthfully it doesn't have the local anymore,
        # using ttl time, stale db record extant.
        self.assertFalse(lock.locked)

        lock2 = MongoLock(self.collection, 'test2')
        # New lock acquire will take ownership based on old ttl\
        self.assertTrue(lock2.acquire())

        # Releasing the original doesn't change ownership
        self.assertTrue(lock.release())
        records = list(self.collection.find())
        self.assertEqual(len(records), 1)

        self.assertEqual(records.pop()['client_id'], lock2.client_id)
        self.assertFalse(lock.acquire(wait=False))

        lock2.release()
        self.assertFalse(list(self.collection.find()))


class QueueTest(TestCase):

    def setUp(self):
        self.client = pymongo.MongoClient(os.environ.get("MONGO_URI"), retryWrites=False, maxPoolSize=1, directConnection=True)
        self.db = self.client.test_queue
        self.queue = Queue(self.db.queue_1, "consumer_1")

    def tearDown(self):
        self.client.drop_database("test_queue")

    def assert_job_equal(self, job, data):
        for k, v in data.items():
            self.assertEqual(job.payload[k], v)

    def test_put_default_channel_next(self):
        data = {"context_id": "pratham",
                "data": [1, 2, 3],
                "more-data": time.time()}
        self.queue.put(dict(data))
        job = self.queue.next()
        self.assert_job_equal(job, data)

    def test_put_custom_channel_next(self):
        data = {"context_id": "pratham",
                "data": [1, 2, 3],
                "more-data": time.time()}
        self.queue.put(dict(data), channel="channle_1", priority=2)
        self.queue.put(dict(data), channel="channle_2", priority=1)
        job = self.queue.next(channel="channle_1")
        self.assert_job_equal(job, data)

    def test_get_empty_queue(self):
        job = self.queue.next()
        self.assertEqual(job, None)

    def test_priority(self):
        self.queue.put({"name": "ek"}, priority=1)
        self.queue.put({"name": "be"}, priority=2)
        self.queue.put({"name": "tran"}, priority=0)

        self.assertEqual(
            ["be", "ek", "tran"],
            [self.queue.next().payload['name'],
             self.queue.next().payload['name'],
             self.queue.next().payload['name']])

    def test_depends_on(self):
        job_id_1 = self.queue.put({"name": "char"})
        job_id_2 = self.queue.put({"name": "panch"})
        self.queue.put({"name": "chh"}, depends_on=[job_id_1])
        self.queue.put({"name": "sat"}, depends_on=[job_id_1, job_id_2])
        next_1 = self.queue.next()
        next_2 = self.queue.next()
        self.assertEqual(["char", "panch"], [next_1.payload['name'], next_2.payload['name']])
        next_1.complete()
        next_3 = self.queue.next()
        self.assertEqual("chh", next_3.payload["name"])
        next_2.complete()
        next_4 = self.queue.next()
        self.assertEqual("sat", next_4.payload["name"])

    def test_complete(self):
        data = {"context_id": "alpha",
                "data": [1, 2, 3],
                "more-data": datetime.now()}

        self.queue.put(data)
        self.assertEqual(self.queue.size(), 1)
        job = self.queue.next()
        job.complete()
        self.assertEqual(self.queue.size(), 0)

    def test_find_locked(self):
        data = {"context_id": "alpha",
                "data": [1, 2, 3]}

        self.queue.put(data)
        self.assertEqual(self.queue.size(), 1)
        job = self.queue.next()
        find_copy_job = self.queue.find_job_by_id(job.job_id)        
        self.assert_job_equal(find_copy_job, data)

    def test_release(self):
        data = {"context_id": "alpha",
                "data": [1, 2, 3],
                "more-data": time.time()}

        self.queue.put(data)
        job = self.queue.next()
        job.release()
        self.assertEqual(self.queue.size(), 1)
        job = self.queue.next()
        self.assert_job_equal(job, data)
        self.assertEqual(job.attempts, 1)
        job.complete()

        self.queue.put(data)
        job = self.queue.next()
        job.release(inc_attempt=False)
        job = self.queue.next()
        # Test if attempt is not being incremented
        self.assertEqual(job.attempts, 0)

    def test_release_state(self):
        data = {"context_id": "alpha",
                "data": [1, 2, 3],
                "more-data": time.time()}
        state = {"current": "state"}

        self.queue.put(data)
        job = self.queue.next()
        job.release(state=state)        
        job = self.queue.next()
        self.assertEqual(job.state, state)

    def test_release_sleep(self):
        data = {"context_id": "alpha",
                "data": [1, 2, 3],
                "more-data": time.time()}        

        self.queue.put(data)
        job = self.queue.next()
        job.release(sleep=10)        
        job = self.queue.next()
        self.assertIsNone(job, "Job is not none, even after sleep of 10 was provided and the next executed immediately after")
        time.sleep(10)
        job = self.queue.next()
        self.assert_job_equal(job, data)

    def test_attempts(self):
        data = {"context_id": "alpha",
                "data": [1, 2, 3],
                "more-data": time.time()}

        self.queue.put(data, channel="attempt_test")
        job = self.queue.next(channel="attempt_test")
        job.release()
        job = self.queue.next(channel="attempt_test")
        job.release()
        job = self.queue.next(channel="attempt_test")
        self.assertEqual(job.attempts, 2)

    def test_error(self):
        data = {"context_id": "alpha",
                "data": [1, 2, 3],
                "more-data": time.time()}

        self.queue.put(data)
        job = self.queue.next()
        job.error("Test Error")
        self.assertEqual(job.last_error, "Test Error")

    def test_ensure_indexes(self):
        indexes = []
        for index in self.queue.collection.list_indexes():
            indexes.append(index.to_dict()["name"])
        required_indexes = ["next_index", "update_index"]
        for index in required_indexes:
            self.assertIn(index, indexes)

    def test_progress(self):
        data = {"context_id": "alpha",
                "data": [1, 2, 3],
                "more-data": time.time()}

        self.queue.put(data)
        job = self.queue.next()
        job.progress(10)
        self.assertEqual(job.progress_count, 10)

    def test_unique_job(self):
        data = {"context_id": "alpha",
                "data": [1, 2, 3],
                "more-data": time.time()}

        self.queue.put(data, job_id=1)
        job2 = self.queue.put({"context_id": "beta",
                "data": [1, 2, 3],
                "more-data": time.time()}, job_id=1)
        self.assertEqual(job2, False)
        self.assert_job_equal(self.queue.next(), data)
